@@ -1,15 +1,18 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+import mimetypes
+
 from sql import Column
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelView, fields
+from trytond.model import ModelView, Workflow, fields
 from trytond.model.exceptions import AccessError
-from trytond.pool import PoolMeta, Pool
-from trytond.pyson import Eval, Bool, Id
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Eval
+from trytond.report import Report
 from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateTransition
+from trytond.wizard import StateAction, StateTransition, Wizard
 
 from .exceptions import PackWarning
 
@@ -21,89 +24,20 @@ else:
     file_id = store_prefix = None
 
 
-class DimensionsMixin(object):
-    __slots__ = ()
-
-    length = fields.Float('Length', digits=(16, Eval('length_digits', 2)),
-        depends=['length_digits'])
-    length_uom = fields.Many2One('product.uom', 'Length Unit',
-        domain=[
-            ('category', '=', Id('product', 'uom_cat_length')),
-            ],
-        states={
-            'required': Bool(Eval('length')),
-            },
-        depends=['length', 'length_digits'])
-    length_digits = fields.Function(fields.Integer('Length Digits'),
-        'on_change_with_length_digits')
-    height = fields.Float('Height', digits=(16, Eval('height_digits', 2)),
-        depends=['height_digits'])
-    height_uom = fields.Many2One('product.uom', 'Height Unit',
-        domain=[
-            ('category', '=', Id('product', 'uom_cat_length')),
-            ],
-        states={
-            'required': Bool(Eval('height')),
-            },
-        depends=['height', 'height_digits'])
-    height_digits = fields.Function(fields.Integer('Height Digits'),
-        'on_change_with_height_digits')
-    width = fields.Float('Width', digits=(16, Eval('width_digits', 2)),
-        depends=['width_digits'])
-    width_uom = fields.Many2One('product.uom', 'Width Unit',
-        domain=[
-            ('category', '=', Id('product', 'uom_cat_length')),
-            ],
-        states={
-            'required': Bool(Eval('width')),
-            },
-        depends=['width', 'width_digits'])
-    width_digits = fields.Function(fields.Integer('Width Digits'),
-        'on_change_with_width_digits')
-
-    @fields.depends('length_uom')
-    def on_change_with_length_digits(self, name=None):
-        return (self.length_uom.digits if self.length_uom
-            else self.default_length_digits())
-
-    @fields.depends('height_uom')
-    def on_change_with_height_digits(self, name=None):
-        return (self.height_uom.digits if self.height_uom
-            else self.default_height_digits())
-
-    @fields.depends('width_uom')
-    def on_change_with_width_digits(self, name=None):
-        return (self.width_uom.digits if self.width_uom
-            else self.default_width_digits())
-
-    @classmethod
-    def default_length_digits(cls):
-        return 2
-
-    @classmethod
-    def default_height_digits(cls):
-        return 2
-
-    @classmethod
-    def default_width_digits(cls):
-        return 2
-
-
-class PackageType(DimensionsMixin, metaclass=PoolMeta):
-    __name__ = 'stock.package.type'
-
-
-class Package(DimensionsMixin, metaclass=PoolMeta):
+class Package(metaclass=PoolMeta):
     __name__ = 'stock.package'
 
     shipping_reference = fields.Char('Shipping Reference',
         states={
-            'readonly': Eval('_parent_shipment', {}).get('carrier', False),
+            'readonly': Eval('has_shipping_service', False),
             })
     shipping_label = fields.Binary(
         "Shipping Label", readonly=True,
         file_id=file_id, store_prefix=store_prefix)
-    shipping_label_id = fields.Char("Shipping Label ID", readonly=True)
+    shipping_label_id = fields.Char(
+        "Shipping Label ID", readonly=True, strip=False)
+    shipping_label_mimetype = fields.Char(
+        "Shipping Label MIME Type", readonly=True)
     shipping_tracking_url = fields.Function(
         fields.Char(
             "Shipping Tracking URL",
@@ -111,10 +45,23 @@ class Package(DimensionsMixin, metaclass=PoolMeta):
                 'invisible': ~Eval('shipping_tracking_url'),
                 }),
         'get_shipping_tracking_url')
+    has_shipping_service = fields.Function(
+        fields.Boolean("Has Shipping Service"),
+        'on_change_with_has_shipping_service')
 
     def get_shipping_tracking_url(self, name):
         return
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.shipping_reference.search_unaccented = False
+        cls._buttons.update(
+            print_shipping_label={
+                'invisible': ~Eval('shipping_label'),
+                'depends': ['shipping_label'],
+                })
+
     @classmethod
     def __register__(cls, module_name):
         pool = Pool()
@@ -145,6 +92,13 @@ class Package(DimensionsMixin, metaclass=PoolMeta):
                     columns=columns,
                     values=values))
 
+    @fields.depends('shipment')
+    def on_change_with_has_shipping_service(self, name=None):
+        return bool(
+            self.shipment
+            and getattr(self.shipment, 'carrier', None)
+            and getattr(self.shipment.carrier, 'shipping_service', None))
+
     @classmethod
     def search_rec_name(cls, name, clause):
         _, operator, value = clause
@@ -158,17 +112,6 @@ class Package(DimensionsMixin, metaclass=PoolMeta):
             ('shipping_reference', *clause[1:]),
             ]
 
-    @fields.depends('type')
-    def on_change_type(self):
-        super().on_change_type()
-        if self.type:
-            self.length = self.type.length
-            self.length_uom = self.type.length_uom
-            self.height = self.type.height
-            self.height_uom = self.type.height_uom
-            self.width = self.type.width
-            self.width_uom = self.type.width_uom
-
     @classmethod
     def copy(cls, packages, default=None):
         if default is None:
@@ -179,31 +122,68 @@ class Package(DimensionsMixin, metaclass=PoolMeta):
         default.setdefault('shipping_label', None)
         return super(Package, cls).copy(packages, default=default)
 
+    @classmethod
+    @ModelView.button_action('stock_package_shipping.report_shipping_label')
+    def print_shipping_label(cls, packages):
+        pass
+
 
 class ShippingMixin:
+    __slots__ = ()
 
+    shipping_reference = fields.Char(
+        "Shipping Reference",
+        states={
+            'readonly': Eval('has_shipping_service', False),
+            })
     shipping_description = fields.Char('Shipping Description',
         states={
             'readonly': Eval('state').in_(['done', 'packed'])
-            },
-        depends=['state'])
+            })
+    has_shipping_service = fields.Function(
+        fields.Boolean("Has Shipping Service"),
+        'on_change_with_has_shipping_service')
 
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        cls.shipping_reference.search_unaccented = False
         cls._buttons.update({
                 'create_shipping': {
-                    'invisible': (Eval('reference', False)
+                    'invisible': (Eval('shipping_reference', False)
                         | ~Eval('carrier', False)),
-                    'readonly': (Eval('reference', False)
+                    'readonly': (Eval('shipping_reference', False)
                         | ~Eval('root_packages', False)
                         | ~Eval('carrier', False)
                         | ~Eval('state').in_(['packed', 'done'])),
-                    'depends': ['state', 'carrier', 'reference',
+                    'depends': [
+                        'state', 'carrier', 'shipping_reference',
                         'root_packages'],
                     },
                 })
 
+    @classmethod
+    def __register__(cls, module):
+        table = cls.__table__()
+        table_h = cls.__table_handler__(module)
+        cursor = Transaction().connection.cursor()
+
+        fill_shiping_reference = (
+            table_h.column_exist('reference')
+            and not table_h.column_exist('shipping_reference'))
+
+        super().__register__(module)
+
+        # Migration from 6.8: fill shipping_reference
+        if fill_shiping_reference:
+            cursor.execute(*table.update(
+                    [table.shipping_reference],
+                    [table.reference]))
+
+    @fields.depends('carrier')
+    def on_change_with_has_shipping_service(self, name=None):
+        return bool(self.carrier and self.carrier.shipping_service)
+
     @classmethod
     def search_rec_name(cls, name, clause):
         _, operator, value = clause
@@ -214,7 +194,7 @@ class ShippingMixin:
         domain = super().search_rec_name(name, clause)
         return [bool_op,
             domain,
-            ('reference', *clause[1:]),
+            ('shipping_reference', *clause[1:]),
             ]
 
     @classmethod
@@ -232,7 +212,10 @@ class ShippingMixin:
         pool = Pool()
         Warning = pool.get('res.user.warning')
         for shipment in shipments:
-            if not shipment.carrier:
+            if (not shipment.carrier
+                    and shipment.delivery_address
+                    and shipment.warehouse not in
+                    shipment.delivery_address.warehouses):
                 name = 'shipment_out_no_carrier_%s' % shipment
                 if Warning.check(name):
                     raise PackWarning(name,
@@ -270,23 +253,6 @@ class ShippingMixin:
 class ShipmentOut(ShippingMixin, metaclass=PoolMeta):
     __name__ = 'stock.shipment.out'
 
-    @classmethod
-    def __register__(cls, module):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        model_data = ModelData.__table__()
-        cursor = Transaction().connection.cursor()
-        super(ShipmentOut, cls).__register__(module)
-
-        # Migration from 4.6: rename create_shipping xml id
-        if module == 'stock_package_shipping':
-            cursor.execute(*model_data.update(
-                    [model_data.fs_id],
-                    ['shipment_out_create_shipping_button'],
-                    where=(model_data.module == module)
-                    & (model_data.model == 'ir.model.button')
-                    & (model_data.fs_id == 'create_shipping_button')))
-
     @classmethod
     @ModelView.button
     @Workflow.transition('packed')
@@ -318,8 +284,7 @@ class ShipmentInReturn(ShippingMixin, metaclass=PoolMeta):
         'carrier', "Carrier",
         states={
             'readonly': ~Eval('state').in_(['draft', 'waiting', 'assigned']),
-            },
-        depends=['state'])
+            })
 
     @classmethod
     @ModelView.button
@@ -356,3 +321,39 @@ class CreateShipping(Wizard):
         method_name = 'validate_packing_%s' % shipping_service
         getattr(self.record, method_name)()
         return 'end'
+
+
+class ShippingLabel(Report):
+    __name__ = 'stock.package.shipping_label'
+
+    @classmethod
+    def render(cls, report, report_context):
+        package = report_context['record']
+        if not package:
+            return '.bin', b''
+        extension = mimetypes.guess_extension(
+            package.shipping_label_mimetype or 'application/octet-stream')
+        # Return with extension so convert has it
+        return extension, package.shipping_label or b''
+
+    @classmethod
+    def convert(cls, report, data, **kwargs):
+        return data
+
+
+class PrintShippingLabel(Wizard):
+    "Print Shipping Label"
+    __name__ = 'stock.shipment.print_shipping_label'
+    start_state = 'print_'
+    print_ = StateAction('stock_package_shipping.report_shipping_label')
+
+    def do_print_(self, action):
+        package_ids = []
+        labels = set()
+        for shipment in self.records:
+            for package in shipment.packages:
+                if (package.shipping_label
+                        and package.shipping_label not in labels):
+                    package_ids.append(package.id)
+                    labels.add(package.shipping_label)
+        return action, {'ids': package_ids}
